unit FBMetaData;

interface

uses
  // general uses
  Data.DB, System.Classes,

  // FireDAC uses
  uADCompClient,

  // IBX uses
  IBDatabase, IBQuery;

type
  TTriggerType = (ttAll, ttBeforeInsert, ttAfterInsert, ttBeforeUpdate,
      ttAfterUpdate, ttBeforeDelete, ttAfterDelete);

  TFBMetaData = class
  private
    FDataBase: TCustomConnection;
    FQuery: TDataSet;
  protected
    procedure Iterator(SQL: string; SL: TStrings); overload;
    procedure Iterator(SQL: string; SL1, SL2: TStrings); overload;
    procedure SetQuery; virtual; abstract;
    function AssignSQL(SQL: string): Boolean; virtual; abstract;
    function ExecSQL(SQL: string): Boolean; virtual;
  public
    constructor Create(DataBase: TCustomConnection); virtual;
    destructor Destroy; override;

    procedure GetTables(TablesName: TStrings);
    procedure GetViews(ViewsName: TStrings);
    procedure GetViewSource(ViewName: string; ViewSource: TStrings);
    procedure GetTableFields(TableName: string; FieldsName: TStrings); overload;
    procedure GetTableFields(TableName: string; FieldsName, FieldsType: TStrings); overload;
    procedure GetPrimaryKeyFields(TableName: string; FieldsName: TStrings); overload;
    procedure GetPrimaryKeyFields(TableName: string; FieldsName, FieldsType: TStrings); overload;
    procedure GetDependOnObject(ObjectName: string; DependObjects, ObjectsType: TStrings);
    procedure GetObjectDependsOn(ObjectName: string; DependObjects, ObjectsType: TStrings);
    procedure GetExceptions(ExceptName, ExceptMessage: TStrings);
    procedure GetFieldArrayDimension(TableName, FieldName: string; LowerBound, UpperBound: TStrings);
    procedure GetUDF(UDF: TStrings);
    procedure GetGenerators(Generators: TStrings);
    procedure GetIndexFields(IndexName: string; IndexFields: TStrings);
    procedure GetIndex(TableName: string; IndexName: TStrings; OnlyActive: Boolean = False; Unique: Boolean = False);
    procedure GetProcedures(ProcNames: TStrings);
    procedure GetProcedureSource(ProcName: string; ProcSource: TStrings);
    procedure GetProcedureInParam(ProcName: string; ProcInParam: TStrings);
    procedure GetProcedureOutParam(ProcName: string; ProcOutParam: TStrings);
    procedure GetRoles(RolesName: TStrings; NoSystemRoles: Boolean = True);
    procedure GetTriggers(TableName: string; TriggerType: TTriggerType; TriggerName: TStrings);
    procedure GetTriggerSource(TriggerName: string; TriggerSource: TStrings);
  end;

  TFBMDFireDAC = class(TFBMetaData)
  protected
    procedure SetQuery; override;
    function AssignSQL(SQL: string): Boolean; override;
  public
    constructor Create(DataBase: TADConnection); reintroduce; virtual;
  end;

  TFBMDIBX = class(TFBMetaData)
  protected
    procedure SetQuery; override;
    function AssignSQL(SQL: string): Boolean; override;
  public
    constructor Create(DataBase: TIBDatabase); reintroduce; virtual;
  end;

implementation

uses
  System.SysUtils;

{ TFBMetaData }

constructor TFBMetaData.Create(DataBase: TCustomConnection);
begin
  FDataBase := DataBase;
  SetQuery;
end;

destructor TFBMetaData.Destroy;
begin
  if Assigned(FQuery) then FreeAndNil(FQuery);

  inherited;
end;

function TFBMetaData.ExecSQL(SQL: string): Boolean;
begin
  Result := False;
  if not Assigned(FQuery) then Exit;

  if FQuery.Active then FQuery.Close;
  if not AssignSQL(SQL) then Exit;
  try
    FQuery.Open;
    Result := True;
  except
    on E: Exception do raise; // you can do your control of this exception
  end;
end;

procedure TFBMetaData.GetTableFields(TableName: string;
  FieldsName: TStrings);
const
  SQL =
    'select rdb$field_name ' +
    'from rdb$relation_fields ' +
    'where upper(rdb$relation_name) = upper(''%s'') and rdb$system_flag = 0 ' +
    'order by rdb$field_position';
begin
  Iterator(Format(SQL, [TableName]), FieldsName);
end;

procedure TFBMetaData.GetPrimaryKeyFields(TableName: string;
  FieldsName: TStrings);
const
  SQL =
    'select s.rdb$field_name ' +
    'from rdb$relation_constraints c ' +
    '  inner join rdb$index_segments s on s.rdb$index_name = c.rdb$index_name ' +
    'where ' +
    '  upper(c.rdb$relation_name) = upper(''%s'') and ' +
    '  c.rdb$constraint_type = ''PRIMARY KEY'' ' +
    'order by s.rdb$field_position';
begin
  Iterator(Format(SQL, [TableName]), FieldsName);
end;

procedure TFBMetaData.GetDependOnObject(ObjectName: string; DependObjects,
  ObjectsType: TStrings);
const
  SQL =
    'select ' +
    '  distinct d.rdb$dependent_name, ' +
    '  case d.rdb$dependent_type ' +
    '    when 0 then ''TABLE'' ' +
    '    when 1 then ''VIEW'' ' +
    '    when 2 then ''TRIGGER'' ' +
    '    when 3 then ''COMPUTED_FIELD'' ' +
    '    when 4 then ''VALIDATION'' ' +
    '    when 5 then ''PROCEDURE'' ' +
    '    when 6 then ''EXPRESSION_INDEX'' ' +
    '    when 7 then ''EXCEPTION'' ' +
    '    when 8 then ''USER'' ' +
    '    when 9 then ''FIELD'' ' +
    '    when 10 then ''INDEX'' ' +
    '    else ''UNDEFINED'' ' +
    '  end rdb$dependent_type ' +
    'from rdb$dependencies d ' +
    'where upper(d.rdb$depended_on_name) = upper(''%s'') ';
begin
  Iterator(Format(SQL, [ObjectName]), DependObjects, ObjectsType);
end;

procedure TFBMetaData.GetExceptions(ExceptName, ExceptMessage: TStrings);
const
  SQL =
    'select e.rdb$exception_name, e.rdb$message from rdb$exceptions e ' +
    'where e.rdb$system_flag = 0 order by e.rdb$exception_number';
begin
  Iterator(SQL, ExceptName, ExceptMessage);
end;

procedure TFBMetaData.GetFieldArrayDimension(TableName, FieldName: string; LowerBound,
  UpperBound: TStrings);
const
  SQL =
    'select d.rdb$lower_bound, d.rdb$upper_bound ' +
    'from rdb$relation_fields f ' +
    '  inner join rdb$field_dimensions d on d.rdb$field_name = f.rdb$field_source ' +
    'where upper(f.rdb$relation_name) = upper(''%s'') and upper(f.rdb$field_name) = upper(''%s'') ' +
    'order by d.rdb$dimension';
begin
  Iterator(Format(SQL, [TableName, FieldName]), LowerBound, UpperBound);
end;

procedure TFBMetaData.GetGenerators(Generators: TStrings);
const
  SQL = 'select g.rdb$generator_name from rdb$generators g where g.rdb$system_flag = 0';
begin
  Iterator(SQL, Generators);
end;

procedure TFBMetaData.GetIndex(TableName: string; IndexName: TStrings;
  OnlyActive, Unique: Boolean);
const
  SQL =
    'select i.rdb$index_name ' +
    'from rdb$indices i ' +
    'where i.rdb$system_flag = 0 and i.rdb$relation_name = ''%s'' %s ' +
    'order by i.rdb$index_name';
  OnlyAc = ' and (i.rdb$index_inactive is null or i.rdb$index_inactive <> 0) ';
  OnlyUni = ' and i.rdb$unique_flag = 1 ';
var
  Tmp: string;
begin
  Tmp := '';
  if OnlyActive then Tmp := OnlyAc;
  if Unique then Tmp := Tmp + OnlyUni;

  Iterator(Format(SQL, [TableName, Tmp]), IndexName);
end;

procedure TFBMetaData.GetIndexFields(IndexName: string;
  IndexFields: TStrings);
const
  SQL =
    'select s.rdb$field_name ' +
    'from rdb$index_segments s ' +
    'where s.rdb$index_name = ''%s'' ' +
    'order by s.rdb$field_position';
begin
  Iterator(Format(SQL, [IndexName]), IndexFields);
end;

procedure TFBMetaData.GetPrimaryKeyFields(TableName: string; FieldsName,
  FieldsType: TStrings);
const
  SQL =
    'select s.rdb$field_name, ' +
    '  case f.rdb$field_type ' +
    '    when 7 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''SMALLINT'' ' +
    '        when 1 then ''NUMERIC('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 8 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''INTEGER'' ' +
    '        when 1 then ''NUMERIC(''  || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 9 then ''QUAD'' ' +
    '    when 10 then ''FLOAT'' ' +
    '    when 12 then ''DATE'' ' +
    '    when 13 then ''TIME'' ' +
    '    when 14 then ''CHAR('' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 16 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''BIGINT'' ' +
    '        when 1 then ''NUMERIC('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 27 then ''DOUBLE'' ' +
    '    when 35 then ''TIMESTAMP'' ' +
    '    when 37 then ''VARCHAR('' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 40 then ''CSTRING'' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 45 then ''BLOB_ID'' ' +
    '    when 261 then ''BLOB SUB_TYPE '' || f.rdb$field_sub_type ' +
    '    else ''RDB$FIELD_TYPE: '' || f.rdb$field_type || ''?'' ' +
    '  end field_type ' +
    'from rdb$index_segments s ' +
    '  join rdb$relation_fields r on r.rdb$field_name = s.rdb$field_name ' +
    '  join rdb$fields f on f.rdb$field_name = r.rdb$field_source ' +
    '  left outer join rdb$character_sets ch on (ch.rdb$character_set_id = f.rdb$character_set_id) ' +
    'where s.rdb$index_name = ( ' +
    '                          select rdb$index_name ' +
    '                          from rdb$relation_constraints ' +
    '                          where upper(rdb$relation_name) = upper(''%s'') and ' +
    '                                rdb$constraint_type = ''PRIMARY KEY'' ' +
    '                         ) and ' +
    '  upper(r.rdb$relation_name) = upper(''%s'') ' +
    'order by s.rdb$field_position';
begin
  Iterator(Format(SQL, [TableName, TableName]), FieldsName, FieldsType);
end;

procedure TFBMetaData.GetProcedureInParam(ProcName: string;
  ProcInParam: TStrings);
const
  SQL =
    'select p.rdb$parameter_name ' +
    'from rdb$procedure_parameters p ' +
    'where p.rdb$procedure_name = ''%s'' and p.rdb$system_flag = 0 and p.rdb$parameter_type = 0 ' +
    'order by p.rdb$parameter_number';
begin
  Iterator(Format(SQL, [ProcName]), ProcInParam);
end;

procedure TFBMetaData.GetProcedureOutParam(ProcName: string;
  ProcOutParam: TStrings);
const
  SQL =
    'select p.rdb$parameter_name ' +
    'from rdb$procedure_parameters p ' +
    'where p.rdb$procedure_name = ''%s'' and p.rdb$system_flag = 0 and p.rdb$parameter_type = 1 ' +
    'order by p.rdb$parameter_number';
begin
  Iterator(Format(SQL, [ProcName]), ProcOutParam);
end;

procedure TFBMetaData.GetProcedures(ProcNames: TStrings);
const
  SQL = 'select p.rdb$procedure_name from rdb$procedures p where p.rdb$system_flag = 0';
begin
  Iterator(SQL, ProcNames);
end;

procedure TFBMetaData.GetProcedureSource(ProcName: string;
  ProcSource: TStrings);
const
  SQL = 'select p.rdb$procedure_source from rdb$procedures p where p.rdb$system_flag = 0 and upper(p.rdb$procedure_name) = upper(''%s'') ';
begin
  if not Assigned(ProcSource) then Exit;
  if not ExecSQL(Format(SQL, [ProcName])) then Exit;
  ProcSource.Text := FQuery.Fields[0].AsString;
end;

procedure TFBMetaData.GetRoles(RolesName: TStrings; NoSystemRoles: Boolean);
const
  SQL = 'select r.rdb$role_name from rdb$roles r %s';
  NoSystem = ' where r.rdb$system_flag = 0';
var
  Tmp: string;
begin
  Tmp := '';
  if NoSystemRoles then Tmp := NoSystem;
  Iterator(Format(SQL, [Tmp]), RolesName);
end;

procedure TFBMetaData.GetObjectDependsOn(ObjectName: string; DependObjects,
  ObjectsType: TStrings);
const
  SQL =
    'select ' +
    '  distinct d.rdb$depended_on_name, ' +
    '  case d.rdb$depended_on_type ' +
    '    when 0 then ''TABLE'' ' +
    '    when 1 then ''VIEW'' ' +
    '    when 2 then ''TRIGGER'' ' +
    '    when 3 then ''COMPUTED_FIELD'' ' +
    '    when 4 then ''VALIDATION'' ' +
    '    when 5 then ''PROCEDURE'' ' +
    '    when 6 then ''EXPRESSION_INDEX'' ' +
    '    when 7 then ''EXCEPTION'' ' +
    '    when 8 then ''USER'' ' +
    '    when 9 then ''FIELD'' ' +
    '    when 10 then ''INDEX'' ' +
    '    else ''UNDEFINED'' ' +
    '  end rdb$depended_on_type ' +
    'from rdb$dependencies d ' +
    'where upper(d.rdb$dependent_name) = upper(''%s'') ';
begin
  Iterator(Format(SQL, [ObjectName]), DependObjects, ObjectsType);
end;

procedure TFBMetaData.GetTableFields(TableName: string; FieldsName,
  FieldsType: TStrings);
const
  SQL =
    'select r.rdb$field_name, ' +
    '  case f.rdb$field_type ' +
    '    when 7 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''SMALLINT'' ' +
    '        when 1 then ''NUMERIC('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 8 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''INTEGER'' ' +
    '        when 1 then ''NUMERIC(''  || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 9 then ''QUAD'' ' +
    '    when 10 then ''FLOAT'' ' +
    '    when 12 then ''DATE'' ' +
    '    when 13 then ''TIME'' ' +
    '    when 14 then ''CHAR('' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 16 then ' +
    '      case f.rdb$field_sub_type ' +
    '        when 0 then ''BIGINT'' ' +
    '        when 1 then ''NUMERIC('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '        when 2 then ''DECIMAL('' || f.rdb$field_precision || '','' || (-f.rdb$field_scale) || '')'' ' +
    '      end ' +
    '    when 27 then ''DOUBLE'' ' +
    '    when 35 then ''TIMESTAMP'' ' +
    '    when 37 then ''VARCHAR('' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 40 then ''CSTRING'' || (trunc(f.rdb$field_length / ch.rdb$bytes_per_character)) || '')'' ' +
    '    when 45 then ''BLOB_ID'' ' +
    '    when 261 then ''BLOB SUB_TYPE '' || f.rdb$field_sub_type ' +
    '    else ''RDB$FIELD_TYPE: '' || f.rdb$field_type || ''?'' ' +
    '  end field_type ' +
    'from rdb$relation_fields r ' +
    '  join rdb$fields f on f.rdb$field_name = r.rdb$field_source ' +
    '  left outer join rdb$character_sets ch on (ch.rdb$character_set_id = f.rdb$character_set_id) ' +
    'where upper(r.rdb$relation_name) = upper(''%s'') ' +
    'order by r.rdb$field_position';
begin
  Iterator(Format(SQL, [TableName]), FieldsName, FieldsType);
end;

procedure TFBMetaData.GetTables(TablesName: TStrings);
const
  SQL =
    'select rdb$relation_name ' +
    'from rdb$relations ' +
    'where rdb$view_blr is null and (rdb$system_flag is null or rdb$system_flag = 0)';
begin
  Iterator(SQL, TablesName);
end;

procedure TFBMetaData.GetTriggers(TableName: string; TriggerType: TTriggerType;
  TriggerName: TStrings);
const
  SQL =
    'select t.rdb$trigger_name ' +
    'from rdb$triggers t ' +
    'where t.rdb$system_flag = 0 and upper(t.rdb$relation_name) = upper(''%s'') %s';
  TType = ' and t.rdb$trigger_type = %d';
var
  Tmp: string;
begin
  case TriggerType of
    ttBeforeInsert: Tmp := Format(TType, [1]);
    ttAfterInsert: Tmp := Format(TType, [2]);
    ttBeforeUpdate: Tmp := Format(TType, [3]);
    ttAfterUpdate: Tmp := Format(TType, [4]);
    ttBeforeDelete: Tmp := Format(TType, [5]);
    ttAfterDelete: Tmp := Format(TType, [6]);
    else Tmp := '';
  end;

  Iterator(Format(SQL, [TableName, Tmp]), TriggerName);
end;

procedure TFBMetaData.GetTriggerSource(TriggerName: string;
  TriggerSource: TStrings);
const
  SQL = 'select t.rdb$trigger_source from rdb$triggers t where t.rdb$system_flag = 0 and upper(t.rdb$trigger_name) = upper(''%s'') ';
begin
  if not Assigned(TriggerSource) then Exit;
  if not ExecSQL(Format(SQL, [TriggerName])) then Exit;
  TriggerSource.Text := FQuery.Fields[0].AsString;
end;

procedure TFBMetaData.GetUDF(UDF: TStrings);
const
  SQL = 'select f.rdb$function_name from rdb$functions f where f.rdb$system_flag = 0';
begin
  Iterator(SQL, UDF);
end;

procedure TFBMetaData.GetViews(ViewsName: TStrings);
const
  SQL =
    'select rdb$relation_name ' +
    'from rdb$relations ' +
    'where rdb$view_blr is not null and (rdb$system_flag is null or rdb$system_flag = 0)';
begin
  Iterator(SQL, ViewsName);
end;

procedure TFBMetaData.GetViewSource(ViewName: string; ViewSource: TStrings);
const
  SQL = 'select r.rdb$view_source from rdb$relations r where r.rdb$relation_name = ''%s'' ';
begin
  if not Assigned(ViewSource) then Exit;
  if not ExecSQL(Format(SQL, [ViewName])) then Exit;
  ViewSource.Text := FQuery.Fields[0].AsString;
end;

procedure TFBMetaData.Iterator(SQL: string; SL: TStrings);
begin
  if not Assigned(SL) then Exit;

  SL.Clear;
  if not ExecSQL(SQL) then Exit;
  while not FQuery.Eof do
  begin
    SL.Add(FQuery.Fields[0].AsString);
    FQuery.Next;
  end;
end;

procedure TFBMetaData.Iterator(SQL: string; SL1, SL2: TStrings);
begin
  if not Assigned(SL1) or not Assigned(SL2) then Exit;

  SL1.Clear;
  SL2.Clear;
  if not ExecSQL(SQL) then Exit;
  while not FQuery.Eof do
  begin
    SL1.Add(FQuery.Fields[0].AsString);
    SL2.Add(FQuery.Fields[1].AsString);
    FQuery.Next;
  end;
end;

{ TFBMDFireDAC }

function TFBMDFireDAC.AssignSQL(SQL: string): Boolean;
begin
  Result := False;
  if not Assigned(FQuery) then Exit; // check assigned
  if not (FQuery is TADQuery) then Exit; // check class

  TADQuery(FQuery).SQL.Text := SQL;
  Result := True;
end;

constructor TFBMDFireDAC.Create(DataBase: TADConnection);
begin
  inherited Create(DataBase);
end;

procedure TFBMDFireDAC.SetQuery;
begin
  FQuery := TADQuery.Create(nil);
  TADQuery(FQuery).Connection := TADConnection(FDataBase);
end;

{ TFBMDIBX }

function TFBMDIBX.AssignSQL(SQL: string): Boolean;
begin
  Result := False;
  if not Assigned(FQuery) then Exit; // check assigned
  if not (FQuery is TIBQuery) then Exit; // check class

  TIBQuery(FQuery).SQL.Text := SQL;
  Result := True;
end;

constructor TFBMDIBX.Create(DataBase: TIBDatabase);
begin
  inherited Create(DataBase);
end;

procedure TFBMDIBX.SetQuery;
begin
  FQuery := TIBQuery.Create(nil);
  TIBQuery(FQuery).Database := TIBDatabase(FDataBase);
end;

end.
